/****************************************************************************\

PSOProxy - http proxy server for the PSO web browser

Copyright (C) 2003 Adrian O'Grady and James Sutherland

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

\****************************************************************************/

// ProcessData.cpp: implementation of the CProcessData class.
//
//////////////////////////////////////////////////////////////////////

#include <windows.h>
#include <stdio.h>
#include <math.h>
#include <png.h>

#include "ProcessData.h"
#include "Settings.h"

static char pDecoderSet[]="TAZOLYNdnE9mP6ci3SzeqIyXBhDgfQp7l5batM4rFKJj8CusxR1+k2V0wUGovWH/";

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CProcessData::CProcessData()
{
	ZeroMemory(pTable, 128);

	int i;

	// Init the decoder mapping table
	char *pSet=pDecoderSet;
    for(i=64; i; --i)
	{
		pTable[*(pSet++)&0x7f]=(64-i);
	};

	float gamma=CSettings::Get().GetValueAsFloat("snapshot", "gamma", 1.0f);

	// Fill in the rgb lookup tables, taking into account the gamma
	for(i=0; i<32; i++)
	{
		float val=i/31.0f;
		val=static_cast<float>(pow(val, 1.0f/gamma));
		pRedLookup[i]=static_cast<unsigned char>(val*255);
		pBlueLookup[i]=static_cast<unsigned char>(val*255);
	}

	// Fill in the rgb lookup tables, taking into account the gamma
	for(i=0; i<64; i++)
	{
		float val=i/63.0f;
		val=static_cast<float>(pow(val, 1.0f/gamma));
		pGreenLookup[i]=static_cast<unsigned char>(val*255);
	}

	pSmoothMask[0].x=-1;
	pSmoothMask[0].y=-1;
	pSmoothMask[0].score=0.5f;
	pSmoothMask[1].x=1;
	pSmoothMask[1].y=1;
	pSmoothMask[1].score=0.5f;
	pSmoothMask[2].x=1;
	pSmoothMask[2].y=-1;
	pSmoothMask[2].score=0.5f;
	pSmoothMask[3].x=-1;
	pSmoothMask[3].y=1;
	pSmoothMask[3].score=0.5f;
	pSmoothMask[4].x=0;
	pSmoothMask[4].y=1;
	pSmoothMask[4].score=1.0f;
	pSmoothMask[5].x=-1;
	pSmoothMask[5].y=0;
	pSmoothMask[5].score=1.0f;
	pSmoothMask[6].x=1;
	pSmoothMask[6].y=0;
	pSmoothMask[6].score=1.0f;
	pSmoothMask[7].x=0;
	pSmoothMask[7].y=-1;
	pSmoothMask[7].score=1.0f;

	m_pFileName[0] = 0;
}

CProcessData::~CProcessData()
{

}

void CProcessData::OnData(const char *Data, unsigned int Length)
{
	// Only know how to deal with one sort of file
	if(Length!=163840) return;

	// Convert data stream back into GC save file
	Decode(const_cast<char*>(Data), pDecodedFile, Length);

	ProcessDecodedData();
}

void CProcessData::OnFile(const char *FileName)
{
	FILE *pFile=fopen(FileName, "rb");
	if(!pFile) return;

	fseek(pFile, 0, SEEK_END);
	int iSize=ftell(pFile);

	if(iSize==(122880))				// Check that it's (most likely) a snapshot
		fseek(pFile, 0, SEEK_SET);
	else if(iSize==(122880+64))		// EMS transfer device adds a 64 byte header
		fseek(pFile, 64, SEEK_SET);
	else
		return;

	fread(pDecodedFile, 1, 122880, pFile);
	fclose(pFile);

	ProcessDecodedData();
}

void CProcessData::ProcessDecodedData()
{
	// Endian swap the whole thing assuming 16 bit data throughout
	for(int i=0; i<GC_FILE_SIZE; i+=2)
	{
		BYTE temp=pDecodedFile[i];
		pDecodedFile[i]=pDecodedFile[i+1];
		pDecodedFile[i+1]=temp;
	}

	GenerateBitmap();

	SYSTEMTIME sysTime;
	GetSystemTime(&sysTime);
	while(sysTime.wYear>=1000)
	{
		sysTime.wYear-=1000;
	}

	char pPath[MAX_PATH];
	strcpy(pPath, CSettings::Get().GetValueAsString("snapshot", "dir", "snapshots\\"));
	int len=strlen(pPath);
	if(pPath[len-1]!='\\')
	{
		strcat(pPath, "\\");
	}

	sprintf(m_pFileName, "snapshot_%02d%02d%02d_%02d%02d%02d.png", sysTime.wDay, sysTime.wMonth, sysTime.wYear, sysTime.wHour, sysTime.wMinute, sysTime.wSecond);
	
	strcat(pPath, m_pFileName);

	WritePNG(pPath);
}

void CProcessData::Decode(char *pIn, char *pOut, int count)
{
	char *pPos=pOut;

	for (int i=0; i<count; i+=4)
	{
		DecodeBytes(&pIn[i], pPos);
		pPos+=3;
	}
}

void CProcessData::DecodeBytes(char pIn[4], char pOut[3])
{
	pOut[0]=DecodeByte(pIn[0])<<2 | DecodeByte(pIn[1])>>4;
	pOut[1]=DecodeByte(pIn[1])<<4 | DecodeByte(pIn[2])>>2;
	pOut[2]=DecodeByte(pIn[2])<<6 | DecodeByte(pIn[3]);
}

void CProcessData::GenerateBitmap()
{
	// Just a whole load of stuff to move pixels around to generate an unswizzled 16 bit image
	
	const int OFFSET=8272;

	int x=0;
	int y=0;
	int i;

	for(i=OFFSET; i<GC_FILE_SIZE; i+=8)
	{
		SetCol16(x, y, ((SHORT*)pDecodedFile)[i/2]);
		x++;
		if(x==256)
		{
			x=0;y+=4;
		}
	}

	x=0;
	y=1;

	for(i=OFFSET+2; i<GC_FILE_SIZE; i+=8)
	{
		SetCol16(x, y, ((SHORT*)pDecodedFile)[i/2]);
		x++;
		if(x==256)
		{
			x=0;y+=4;
		}
	}

	x=0;
	y=2;

	for(i=OFFSET+4; i<GC_FILE_SIZE; i+=8)
	{
		SetCol16(x, y, ((SHORT*)pDecodedFile)[i/2]);
		x++;
		if(x==256)
		{
			x=0;y+=4;
		}
	}

	x=0;
	y=3;

	for(i=OFFSET+6; i<GC_FILE_SIZE; i+=8)
	{
		SetCol16(x, y, ((SHORT*)pDecodedFile)[i/2]);
		x++;
		if(x==256)
		{
			x=0;y+=4;
		}
	}

	for(x=0; x<256; x+=4)
	{
		for(y=0; y<192; y+=4)
		{
			SHORT v[4][4];
			v[0][0]=GetCol16(0+x,0+y);
			v[1][0]=GetCol16(1+x,0+y);
			v[2][0]=GetCol16(2+x,0+y);
			v[3][0]=GetCol16(3+x,0+y);
			v[0][1]=GetCol16(0+x,1+y);
			v[1][1]=GetCol16(1+x,1+y);
			v[2][1]=GetCol16(2+x,1+y);
			v[3][1]=GetCol16(3+x,1+y);
			v[0][2]=GetCol16(0+x,2+y);
			v[1][2]=GetCol16(1+x,2+y);
			v[2][2]=GetCol16(2+x,2+y);
			v[3][2]=GetCol16(3+x,2+y);
			v[0][3]=GetCol16(0+x,3+y);
			v[1][3]=GetCol16(1+x,3+y);
			v[2][3]=GetCol16(2+x,3+y);
			v[3][3]=GetCol16(3+x,3+y);

			SetCol16(0+x,0+y,v[0][0]);
			SetCol16(1+x,0+y,v[0][1]);
			SetCol16(2+x,0+y,v[0][2]);
			SetCol16(3+x,0+y,v[0][3]);
			SetCol16(0+x,1+y,v[1][0]);
			SetCol16(1+x,1+y,v[1][1]);
			SetCol16(2+x,1+y,v[1][2]);
			SetCol16(3+x,1+y,v[1][3]);
			SetCol16(0+x,2+y,v[2][0]);
			SetCol16(1+x,2+y,v[2][1]);
			SetCol16(2+x,2+y,v[2][2]);
			SetCol16(3+x,2+y,v[2][3]);
			SetCol16(0+x,3+y,v[3][0]);
			SetCol16(1+x,3+y,v[3][1]);
			SetCol16(2+x,3+y,v[3][2]);
			SetCol16(3+x,3+y,v[3][3]);
		}
	}
}

// Old function.  To be removed.
void CProcessData::WriteBMP(const char *pFileName)
{
	// Generate the 24 bit image
	int pos=0;

	for(int y=0; y<192; y++)
	{
		for(int x=0; x<256; x++)
		{
			int r, g, b;

			r=(GetCol16(x, 191-y)>>11)&0x1f;
			r=static_cast<int>(r*(255.0f/31.0f));

			g=(GetCol16(x, 191-y)>>5)&0x3f;
			g=static_cast<int>(g*(255.0f/63.0f));

			b=(GetCol16(x, 191-y))&0x1f;
			b=static_cast<int>(b*(255.0f/31.0f));

			pBitmap24[pos++]=b;
			pBitmap24[pos++]=g;
			pBitmap24[pos++]=r;
		}
	}

	// Generate BMP headers
	BITMAPFILEHEADER bmfh;
	ZeroMemory(&bmfh, sizeof(BITMAPFILEHEADER));

	bmfh.bfType=19778;
	bmfh.bfOffBits=sizeof(BITMAPFILEHEADER)+sizeof(BITMAPINFOHEADER);
	bmfh.bfSize=bmfh.bfOffBits+(256*192*3);

	BITMAPINFOHEADER bmi;
	ZeroMemory(&bmi, sizeof(BITMAPINFOHEADER));
	bmi.biSize=sizeof(BITMAPINFOHEADER);
	bmi.biWidth=256;
	bmi.biHeight=192;
	bmi.biPlanes=1;
	bmi.biBitCount=24;

	// Write out file
	FILE* bmp=fopen(pFileName, "wb");
	fwrite(&bmfh, sizeof(BITMAPFILEHEADER), 1, bmp);
	fwrite(&bmi, sizeof(BITMAPINFOHEADER), 1, bmp);
	fwrite(pBitmap24, 1, 256*192*3, bmp);
	fclose(bmp);
}

/*
int Jitter(int val, int prev, int next, float noise)
{
	int low, high;

	low=val-static_cast<int>((val-prev)*noise);
	high=val+static_cast<int>((next-val)*noise);

	if(high==low) return val;
	return (rand()%(high-low))+low;
}

int Tween(int base, int dest, float alpha)
{
	return static_cast<int>(static_cast<float>(dest)*alpha+static_cast<float>(base)*(1.0f-alpha));
}
*/
/*
// Work in progress
void CProcessData::Generate24BitValueNice(int x, int y, int& r, int& g, int& b)
{
	r=(GetCol16(x, y)>>11)&0x1f;
	g=(GetCol16(x, y)>>5)&0x3f;
	b=(GetCol16(x, y))&0x1f;

	float scoreR=0;
	float scoreG=0;
	float scoreB=0;

	for(int i=0; i<8; i++)
	{
		int testR=(GetCol16(pSmoothMask[i].x, pSmoothMask[i].y)>>11)&0x1f;
		int testG=(GetCol16(pSmoothMask[i].x, pSmoothMask[i].y)>>5)&0x3f;
		int testB=(GetCol16(pSmoothMask[i].x, pSmoothMask[i].y))&0x1f;

		if(testR==(r-1)) scoreR-=pSmoothMask[i].score;
		if(testR==(r+1)) scoreR+=pSmoothMask[i].score;
		
		if(testG==(g-1)) scoreG-=pSmoothMask[i].score;
		if(testG==(g+1)) scoreG+=pSmoothMask[i].score;

		if(testB==(b-1)) scoreB-=pSmoothMask[i].score;
		if(testB==(b+1)) scoreB+=pSmoothMask[i].score;
	}

	if(scoreR<0)
	{
		r=Tween(pRedLookup[r], pRedLookup[r>0?r-1:r], -scoreR/6.0f);
	}
	else
	{
		r=Tween(pRedLookup[r], pRedLookup[r<31?r+1:r], scoreR/6.0f);
	}

	if(scoreG<0)
	{
		g=Tween(pGreenLookup[g], pGreenLookup[g>0?g-1:g], -scoreG/6.0f);
	}
	else
	{
		g=Tween(pGreenLookup[g], pGreenLookup[g<63?g+1:g], scoreG/6.0f);
	}

	if(scoreB<0)
	{
		b=Tween(pBlueLookup[b], pBlueLookup[b>0?b-1:b], -scoreB/6.0f);
	}
	else
	{
		b=Tween(pBlueLookup[b], pBlueLookup[b<31?b+1:b], scoreB/6.0f);
	}
}
*/

void CProcessData::Generate24BitValue(int x, int y, int sharp, int& r, int& g, int& b)
{
	r=pRedLookup[(GetCol16(x, y)>>11)&0x1f];
	g=pGreenLookup[(GetCol16(x, y)>>5)&0x3f];
	b=pBlueLookup[(GetCol16(x, y))&0x1f];

	if(!sharp) return;

	int rUp, gUp, bUp;
	int rDown, gDown, bDown;

	rUp=pRedLookup[(GetCol16(x, y>0?y-1:y)>>11)&0x1f];
	gUp=pGreenLookup[(GetCol16(x, y>0?y-1:y)>>5)&0x3f];
	bUp=pBlueLookup[(GetCol16(x, y>0?y-1:y))&0x1f];

	rDown=pRedLookup[(GetCol16(x, y<191?y+1:y)>>11)&0x1f];
	gDown=pGreenLookup[(GetCol16(x, y<191?y+1:y)>>5)&0x3f];
	bDown=pBlueLookup[(GetCol16(x, y<191?y+1:y))&0x1f];

	if(sharp==1)
	{
		r=((r*4)-rUp-rDown)/2;
		g=((g*4)-gUp-gDown)/2;
		b=((b*4)-bUp-bDown)/2;
	}
	else
	{
		r=((r*3)-rUp-rDown);
		g=((g*3)-gUp-gDown);
		b=((b*3)-bUp-bDown);
	}

	if(r<0) r=0; if(r>255) r=255;
	if(g<0) g=0; if(g>255) g=255;
	if(b<0) b=0; if(b>255) b=255;
}

void CProcessData::WritePNG(const char *pFileName)
{
	int i;

	FILE *fp = fopen(pFileName, "wb");
	if (!fp)
	{
		return;
	}

    png_structp png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    if (!png_ptr)
       return;

    png_infop info_ptr = png_create_info_struct(png_ptr);
    if(!info_ptr)
    {
       png_destroy_write_struct(&png_ptr, (png_infopp)NULL);
       return;
    }

	png_init_io(png_ptr, fp);

	png_set_IHDR(	png_ptr, info_ptr, 256, 192,
					8, PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE ,
					PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);

	png_bytep row_pointers[192];
	for(i=0; i<192; i++)
	{
		row_pointers[i]=reinterpret_cast<png_bytep>(png_malloc(png_ptr, 256*3));
	}

	// Generate the 24 bit image
	int pos=0;

	int sharp=CSettings::Get().GetValueAsInt("snapshot", "sharpness", 1);

	for(int y=0; y<192; y++)
	{
		for(int x=0; x<256; x++)
		{
			int r, g, b;

			Generate24BitValue(x, y, sharp, r, g, b);

			row_pointers[y][x*3]=r;
			row_pointers[y][(x*3)+1]=g;
			row_pointers[y][(x*3)+2]=b;
		}
	}

	png_set_rows(png_ptr, info_ptr, row_pointers);

    png_write_png(png_ptr, info_ptr, 0, NULL);

	for(i=0; i<192; i++)
	{
		png_free(png_ptr, row_pointers[i]);
	}

	png_destroy_write_struct(&png_ptr, &info_ptr);

	fclose(fp);
}