/****************************************************************************\

PSOProxy - http proxy server for the PSO web browser

Copyright (C) 2003 Adrian O'Grady and James Sutherland

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

\****************************************************************************/

// Settings.cpp: implementation of the CSettings class.
//
//////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "Settings.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CSettings	CSettings::s_inst;

CSettings::CSettings()
{
	m_Sections = 0;
	m_Filename[0] = 0;
}

CSettings::~CSettings()
{
	if (m_Sections) DestroySections(m_Sections);
}

bool CSettings::LoadIni(const char *Filename)
{
	char Line[80];
	char * StartData = 0;
	char * EndData = 0;
	sSection * CurrentSection = 0;

	strcpy(m_Filename, Filename);

	ifstream iniFile(Filename);

	while (iniFile.getline(Line, 80))
	{
		switch (CheckLineType(Line))
		{
		case SECTION:
			StartData = Line + 1;
			EndData = strstr(StartData, "]");
			if (EndData)
			{
				*EndData = 0;
				CurrentSection = AddSection(StartData, m_Sections);
			}
			break;

		case DATA:
			StartData = strstr(Line, "=");
			*StartData = 0;
			StartData++;

			if (CurrentSection)
			{
				AddData(Line, StartData, CurrentSection->Data);
			}
			break;

		default:
			// blank, comment or broken lines
			break;
		}
	}

	iniFile.close();
	m_IsDirty = false;

	return true;
}

CSettings::eLineType CSettings::CheckLineType(const char *Line)
{
	eLineType r = BLANK;

	if (Line[0] == '[')
	{
		r = SECTION;
	}
	else if (Line[0] == '#')
	{
		r = COMMENT;
	}
	else if (strstr(Line, "="))
	{
		r = DATA;
	}

	return r;
}

CSettings::sSection * CSettings::AddSection(const char *Name, sSection *& Head)
{
	if (Head)
	{
		return AddSection(Name, Head->Next);
	}
	else
	{
		Head = new sSection;
		Head->Name = new char[strlen(Name) + 1];
		strcpy(Head->Name, Name);
		Head->Next = 0;
		Head->Data = 0;

		return Head;
	}
}

void CSettings::DestroyData(sKVPair *Head)
{
	if (Head->Next)
	{
		DestroyData(Head->Next);
	}

	delete[] Head->Key;
	delete[] Head->Value;
	delete Head;

	return;
}

void CSettings::DestroySections(sSection *Head)
{
	if (Head->Next)
	{
		DestroySections(Head->Next);
	}

	if (Head->Data)
	{
		DestroyData(Head->Data);
	}

	delete[] Head->Name;
	delete Head;
}

void CSettings::AddData(const char *Key, const char *Value, sKVPair *& Head)
{
	if (Head)
	{
		AddData(Key, Value, Head->Next);
	}
	else
	{
		Head = new sKVPair;
		Head->Next = 0;
		Head->Key = new char[strlen(Key) + 1];
		strcpy(Head->Key, Key);
		Head->Value = new char[strlen(Value) + 1];
		strcpy(Head->Value, Value);
	}
}

const char * CSettings::GetValueAsString(const char *Section, const char *Key)
{
	const char * r = 0;
	sSection * current = FindSection(Section, m_Sections);

	if (current)
	{
		sKVPair * pair = FindKey(Key, current->Data);
		if (pair)
		{
			r = pair->Value;
		}
	}

	return r;
}

int CSettings::GetValueAsInt(const char *Section, const char *Key)
{
	int r = 0;
	const char * value = GetValueAsString(Section, Key);

	if (value)
	{
		atoi(value);
	}

	return r;
}

float CSettings::GetValueAsFloat(const char *Section, const char *Key)
{
	float r = 0;
	const char * value = GetValueAsString(Section, Key);

	if (value)
	{
		sscanf(value, "%f", &r);
	}

	return r;
}

CSettings::sSection * CSettings::FindSection(const char *Name, sSection *Head)
{
	sSection * r = 0;

	if (Head)
	{
		if (strcmp(Head->Name, Name) == 0)
		{
			r = Head;
		}
		else
		{
			r = FindSection(Name, Head->Next);
		}
	}

	return r;
}

CSettings::sKVPair * CSettings::FindKey(const char *Key, sKVPair *Head)
{
	sKVPair * r = 0;

	if (Head)
	{
		if (strcmp(Head->Key, Key) == 0)
		{
			r = Head;
		}
		else
		{
			r = FindKey(Key, Head->Next);
		}
	}

	return r;
}

void CSettings::SaveIni()
{
	if (m_Filename && m_IsDirty)
	{
		ofstream file(m_Filename);

		SaveSection(m_Sections, file);

		file.close();
	}
}

void CSettings::SaveSection(sSection *Section, ofstream &file)
{
	if (Section)
	{
		char Line[80];
		sprintf(Line, "[%s]\n", Section->Name);
		file.write(Line, strlen(Line));

		SaveData(Section->Data, file);
		SaveSection(Section->Next, file);
	}
}

void CSettings::SaveData(sKVPair * Head, ofstream &file)
{
	char Line[80];

	if (Head)
	{
		sprintf(Line, "%s=%s\n", Head->Key, Head->Value);
		file.write(Line, strlen(Line));
		SaveData(Head->Next, file);
	}
	else
	{
		sprintf(Line, "\n");
		file.write(Line, strlen(Line));
	}
}

void CSettings::SetValue(const char *Section, const char *Key, const char *Value)
{
	sSection * sec = FindSection(Section, m_Sections);

	if (sec)
	{
		sKVPair * pair = FindKey(Key, sec->Data);
		if (pair)
		{
			delete[] pair->Value;
			pair->Value = new char[strlen(Value) + 1];
			strcpy(pair->Value, Value);
		}
		else
		{
			AddData(Key, Value, sec->Data);
		}
	}
	else
	{
		sec = AddSection(Section, m_Sections);
		AddData(Key, Value, sec->Data);
	}

	m_IsDirty = true;
}

void CSettings::SetValue(const char *Section, const char *Key, int Value)
{
	char buf[32];
	sprintf(buf, "%d", Value);
	SetValue(Section, Key, buf);
}

void CSettings::SetValue(const char *Section, const char *Key, float Value)
{
	char buf[64];
	sprintf(buf, "%f", Value);
	SetValue(Section, Key, buf);
}

int CSettings::GetValueAsInt(const char *Section, const char *Key, int Default)
{
	int value = 0;
	const char * retval = GetValueAsString(Section, Key);

	if (retval)
	{
		value = atoi(retval);
	}
	else
	{
		value = Default;
		SetValue(Section, Key, Default);
	}

	return value;
}

float CSettings::GetValueAsFloat(const char *Section, const char *Key, float Default)
{
	float r = Default;
	const char * value = GetValueAsString(Section, Key);

	if (value)
	{
		sscanf(value, "%f", &r);
	}
	else
	{
		SetValue(Section, Key, Default);
	}

	return r;
}

const char * CSettings::GetValueAsString(const char *Section, const char *Key, const char *Default)
{
	const char * r = GetValueAsString(Section, Key);
	if (r == 0)
	{
		SetValue(Section, Key, Default);
		r = Default;
	}

	return r;
}

void CSettings::GetAllSettings(ISettingsEnumerator * callback, const char *Section)
{
	sSection * pSection = FindSection(Section, m_Sections);
	EnumerateData(callback, pSection->Data);
}

void CSettings::EnumerateData(ISettingsEnumerator * callback, sKVPair *Pair)
{
	if (Pair)
	{
		callback->OnDataItem(Pair->Key, Pair->Value, false);
		EnumerateData(callback, Pair->Next);
	}
	else
	{
		callback->OnDataItem(0, 0, true);
	}
}
