// ProxyAPI1.cpp: implementation of the CProxyAPI1 class.
//
//////////////////////////////////////////////////////////////////////

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "ProxyAPI1.h"
#include "Settings.h"
#include "SysLog.h"
#include "AppletDB.h"
#include "VirtualPathMan.h"

const char * __GetServerVersionString();
void __GetServerVersion(int& Major, int& Minor, int& Revision);

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

HANDLE CProxyAPI1::m_LogSemaphore = 0;
HANDLE CProxyAPI1::m_SettingsSemaphore = 0;

CProxyAPI1::CProxyAPI1()
{
	m_Parent = 0;
	m_Header = 0;
	m_BufInput = 0;
	m_BufInputLen = 0;
	m_AppName[0] = 0;
	m_Socket = 0;
	m_Owner = 0;
}

CProxyAPI1::~CProxyAPI1()
{
	if (m_BufInput) delete[] m_BufInput;
}

int CProxyAPI1::ExecApplet(const char *AppletName, const char *Args)
{
	return CAppletDB::Get().ExecApplet(AppletName, Args, this);
}

IHttpHeader& CProxyAPI1::GetHttpHeader()
{
	return *m_Header;
}

int CProxyAPI1::GetSettingAsInt(const char *Key, int Default)
{
	int r = 0;

	if (m_AppName)
	{
		WaitForSingleObject(m_SettingsSemaphore, INFINITE);
		r = CSettings::Get().GetValueAsInt(m_AppName, Key, Default);
		ReleaseSemaphore(m_SettingsSemaphore, 1, NULL);
	}

	return r;
}

float CProxyAPI1::GetSettingAsFloat(const char *Key, float Default)
{
	float r = 0;

	if (m_AppName)
	{
		WaitForSingleObject(m_SettingsSemaphore, INFINITE);
		r = CSettings::Get().GetValueAsFloat(m_AppName, Key, Default);
		ReleaseSemaphore(m_SettingsSemaphore, 1, NULL);
	}

	return r;
}

const char * CProxyAPI1::GetSettingAsString(const char *Key, const char *Default)
{
	const char * r = 0;

	if (m_AppName)
	{
		WaitForSingleObject(m_SettingsSemaphore, INFINITE);
		r = CSettings::Get().GetValueAsString(m_AppName, Key, Default);
		ReleaseSemaphore(m_SettingsSemaphore, 1, NULL);
	}

	return r;
}

void CProxyAPI1::Log(const char *Text)
{
	if (m_AppName && Text)
	{
		char * Buffer = new char[strlen(Text) + strlen(m_AppName) + 3];

		sprintf(Buffer, "%s: %s", m_AppName, Text);

		WaitForSingleObject(m_LogSemaphore, INFINITE);
		CSysLog::Get().Log(Buffer);
		ReleaseSemaphore(m_LogSemaphore, 1, NULL);

		delete[] Buffer;
	}
}

int CProxyAPI1::Recv(char *Data, int Length)
{
	int r = SOCKET_ERROR;

	if (m_Parent)
	{
		r = m_Parent->Recv(Data, Length);
	}
	else if (m_BufInputLen != 0)
	{
		if (m_BufInputLen <= Length)
		{
			// Empty the buffer in one swoop
			r = m_BufInputLen;
			memcpy(Data, m_BufInput, m_BufInputLen);
			m_BufInputLen = 0;

		}
		else
		{
			memcpy(Data, m_BufInput, Length);
			memcpy(m_BufInput, &m_BufInput[Length], Length);
			m_BufInputLen -= Length;
			r = Length;
		}

	}
	else
	{
		r = recv(m_Socket, Data, Length, 0);
	}

	return r;
}

int CProxyAPI1::Send(const char *Data, int Length)
{
	int r = SOCKET_ERROR;

	if (m_Parent)
	{
		r = m_Parent->Send(Data, Length);
	}
	else
	{
		r = send(m_Socket, Data, Length, 0);
	}

	return r;
}

void CProxyAPI1::SendHeader(const char *Result, const char *ContentType, unsigned int ContentLength, const char *Extra)
{
	char Buffer[1024];
	char Line[128];
	
	// Build a nice simple header
	sprintf(Line, "HTTP/1.0 %s\x0D\x0A", Result);
	strcpy(Buffer, Line);
	sprintf(Line, "Server: %s\x0D\x0A", __GetServerVersionString());
	strcat(Buffer, Line);
	strcat(Buffer, "Proxy-Connection: close\x0D\x0A");
	if (ContentLength)
	{
		sprintf(Line, "Content-Length: %d\x0D\x0A", ContentLength);
		strcat(Buffer, Line);
	}
	sprintf(Line, "Content-Type: %s\x0D\x0A", ContentType);
	strcat(Buffer, Line);

	Send(Buffer, strlen(Buffer));

	if (Extra)
	{
		Send(Extra, strlen(Extra));
	}

	Send("\x0D\x0A", 2);
}

void CProxyAPI1::SendInternalError(const char *Error)
{
	char Buffer[1024];

	SendHeader("500 Server Error", "text/html", 0, 0);

	sprintf(Buffer, "<HTML><HEAD><TITLE>Internal Error</TITLE></HEAD><BODY>%s</BODY></HTML>", Error);
	Send(Buffer, strlen(Buffer));
}

bool CProxyAPI1::SendPage(const char *Filename)
{
	bool r = false;
	int ReadLength = 0;
	char ReadBuffer[1460];
	FILE * file = fopen(Filename, "rb");

	if (file)
	{
		// Loop until we've sent all the file
		do
		{
			// Read the file
			ReadLength = fread(ReadBuffer, 1, 1460, file);
			if (ReadLength > 0)
			{
				// There was data, send it
				Send(ReadBuffer, ReadLength);
			}
		} while (!feof(file));

		fclose(file);
		r = true;
	}
	
	return r;
}

void CProxyAPI1::SetSetting(const char *Key, const char *Value)
{
	if (m_AppName)
	{
		WaitForSingleObject(m_SettingsSemaphore, INFINITE);
		CSettings::Get().SetValue(m_AppName, Key, Value);
		ReleaseSemaphore(m_SettingsSemaphore, 1, NULL);
	}
}

void CProxyAPI1::SetSetting(const char *Key, int Value)
{
	if (m_AppName)
	{
		WaitForSingleObject(m_SettingsSemaphore, INFINITE);
		CSettings::Get().SetValue(m_AppName, Key, Value);
		ReleaseSemaphore(m_SettingsSemaphore, 1, NULL);
	}
}

void CProxyAPI1::SetSetting(const char *Key, float Value)
{
	if (m_AppName)
	{
		WaitForSingleObject(m_SettingsSemaphore, INFINITE);
		CSettings::Get().SetValue(m_AppName, Key, Value);
		ReleaseSemaphore(m_SettingsSemaphore, 1, NULL);
	}
}

void CProxyAPI1::GetAllSettings(ISettingsEnumerator *callback)
{
	if (m_AppName)
	{
		WaitForSingleObject(m_SettingsSemaphore, INFINITE);
		CSettings::Get().GetAllSettings(callback, m_AppName);
		ReleaseSemaphore(m_SettingsSemaphore, 1, NULL);
	}
}

bool CProxyAPI1::StaticInitialise()
{
	bool r = false;

	if (m_LogSemaphore == 0)
	{
		m_LogSemaphore = CreateSemaphore(NULL, 1, 1, NULL);
		m_SettingsSemaphore = CreateSemaphore(NULL, 1, 1, NULL);

		r = true;
	}

	return r;
}

void CProxyAPI1::StaticDeinitialise()
{
	if (m_LogSemaphore)
	{
		CloseHandle(m_LogSemaphore);
		CloseHandle(m_SettingsSemaphore);
	}
}

const char * CProxyAPI1::GetServerVersionString()
{
	return __GetServerVersionString();
}

void CProxyAPI1::GetServerVersion(int& Major, int& Minor, int& Revision)
{
	__GetServerVersion(Major, Minor, Revision);
}
void CProxyAPI1::SetOwner(IProxyApplet1 *Owner)
{
	m_Owner = Owner;
}

bool CProxyAPI1::SendDynamicPage(const char *Filename)
{
	//char RealFilename[256];
	char Line[256];
	char * Found = 0;
	char Number[16];
	int TagNumber = 0;

	//CVirtualPathMan::Get().GetRealPath(Filename, RealFilename);

	ifstream DynamicPage(Filename);

	while (DynamicPage.getline(Line, 255))
	{
		// Look for the proxy tag and send everything before it
		Found = strstr(Line, "$$");
		if (Found)
		{
			Send(Line, (Found - Line));
			// Move the found point to the characters after the tag
			Found += 2;
			CopyNumber(Found, Number);

			if (Number[0] != 0)
			{
				TagNumber = atoi(Number);
				// Move the found pointer on past the number
				Found += strlen(Number);

				if (m_Owner)
				{
					m_Owner->OnProxyTag(TagNumber);
				}

				// Send what's left of the line
				Send(Found, strlen(Found));
				Send("\x0D\x0A", 2);
			}
			else
			{
				// Didn't get a valid tag, so fix the pointer and just send the text
				Found -= 2;
				Send(Found, strlen(Found));
				Send("\x0D\x0A", 2);
			}
		}
		else
		{
			Send(Line, strlen(Line));
			Send("\x0D\x0A", 2);
		}
	}

	DynamicPage.close();

	return true;
}

char * CProxyAPI1::CopyNumber(const char *Text, char *Number)
{
	bool Quit = false;
	unsigned int x = 0;
	
	while (Quit == false)
	{
		if ((Text[x] >= '0') && (Text[x] <= '9'))
		{
			Number[x] = Text[x];
		}
		else
		{
			Number[x] = 0;
			Quit = true;
		}
		x++;
	}

	return Number;
}
